/* $Workfile: DspCore2Core.c $   $Revision: 5 $    $Modtime: 11/05/09 6:19a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 2007, All Rights Reserved

*******************************************************************************/

// Only include with DaVinci
#ifdef DAVINCI
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#if OSA_DBG
#define DBG 1
#endif
#include "OsaLinuxRtl.h"  // For OSA_DPRINTF
#if OSA_DBG
#undef DBG
#endif
#include "OsaLinuxCallbacks.h"
#include "OsaLinuxAsp.h"
#include "OsaLinuxDmp.h"
#include "DspCore2Core.h"
#include <asm/arch/hardware.h>
#include <asm/arch/edma.h>
#include <asm/arch/mcbsp.h>

#include <Sys/DeviceDaVinci/OdDaVinciNicDef.h>

#ifdef USE_CORE2CORE_SIGNALS_FOR_ASP

//CHNG: declaration
typedef struct {
  u32 dwRxTxFlags;
  u32 coBufPartSize;
  u32 nOffsets[4];
} OtxAspBuffersStateS;

#define CHIP_MEM  0x11104000  // Start of VICP memory (32k long?)
#define OTX_ASP_BUFFERS_BASE_ADDR CHIP_MEM

#define OTX_ASP_BUFFERS_STATE_ADDR ((OtxAspBuffersStateS *)ioremap((OTX_ASP_BUFFERS_BASE_ADDR + (32*1024)), sizeof(OtxAspBuffersStateS))) //TODO: verify offset
#define OTX_ASP_BUFFERS_RX_PING_FLAG (0x000000FF)
#define OTX_ASP_BUFFERS_RX_PONG_FLAG (0x0000FF00)
#define OTX_ASP_BUFFERS_TX_PING_FLAG (0x00FF0000)
#define OTX_ASP_BUFFERS_TX_PONG_FLAG (0xFF000000)

#define DAVINCI_INTGEN 0x1C40010
#define DAVINCI_INTDSP3 7


/* From DSP */
#define OTX_EXTEN_BUFFERS_BASE_ADDR (OTX_ASP_BUFFERS_BASE_ADDR + (40*1024))

#define OTX_EXTEN_RX_BUF_ADDR ((char *)ioremap(OTX_EXTEN_BUFFERS_BASE_ADDR, 8*1024))
#define OTX_EXTEN_TX_BUF_ADDR ((char *)ioremap(OTX_EXTEN_BUFFERS_BASE_ADDR + 8*1024, 8*1024))
/* End of from DSP */

OtxAspBuffersStateS *g_pBuffersState = 0;//OTX_ASP_BUFFERS_STATE_ADDR;
extern int g_cbBufSize;
int   g_DspBufSize;
char* g_RxDspBufPtr;
char* g_TxDspBufPtr;

struct DspParams {
  unsigned int ver;
  void* dspptr;
};

void* DspCore2CoreParams()
{
  static struct DspParams params;
  static int inited = 0;
  if( !inited ) {
    params.ver = 0x01;
    params.dspptr = ioremap( 0x11104000 + 32 * 1024 + 64, 128 );
    inited = 1;
  }
  return &params;
}

int DspCore2CoreRx(int nSourceId, int nBoardType, char * buf, size_t count)
{
  if (OsaLinuxAspPeekBufPtr(0,0)==buf) {
    g_pBuffersState->dwRxTxFlags |= OTX_ASP_BUFFERS_RX_PING_FLAG;
  } else if (OsaLinuxAspPeekBufPtr(0,1)==buf) {
    g_pBuffersState->dwRxTxFlags |= OTX_ASP_BUFFERS_RX_PONG_FLAG;
  } else {
    return -1;
  }
  OsaLinuxDmpWriteReg32(DAVINCI_INTGEN, (1 << DAVINCI_INTDSP3));
  return 0;
}

int DspCore2CoreTx(int nSourceId, int nBoardType, char * buf, size_t count)
{
  if (OsaLinuxAspPeekBufPtr(1,0)==buf) {
    g_pBuffersState->dwRxTxFlags |= OTX_ASP_BUFFERS_TX_PING_FLAG;
  } else if (OsaLinuxAspPeekBufPtr(1,1)==buf) {
    g_pBuffersState->dwRxTxFlags |= OTX_ASP_BUFFERS_TX_PONG_FLAG;
  } else {
    return -1;
  }
  OsaLinuxDmpWriteReg32(DAVINCI_INTGEN, (1 << DAVINCI_INTDSP3));
  return 0;
}

int DspCore2CoreInterruptsEnable(void)
{
  unsigned long p;
  OsaLinuxDmpReadReg32(OD_DMP_BITEN_REG, &p);
  OsaLinuxDmpWriteReg32(OD_DMP_BITEN_REG, p&~1);
  // here we have disabled interrup from the Bank 0
  OsaLinuxDmpReadReg32(OD_DMP_DIR01_REG, &p);
  OsaLinuxDmpWriteReg32(OD_DMP_DIR01_REG, p&~(1<<OTX_CORE2CORE_GPIONO));
  // here we have changed direction for the GPIO 1
  OsaLinuxDmpWriteReg32(OD_DMP_CLR_DATA01_REG, (1<<OTX_CORE2CORE_GPIONO));
  // here we have set pin 1 to low
  OsaLinuxDmpReadReg32(OD_DMP_BITEN_REG, &p);
  OsaLinuxDmpWriteReg32(OD_DMP_BITEN_REG, p|1);
  // here we have enabled interrup from the Bank 0
  OsaLinuxDmpWriteReg32(OD_DMP_SET_RIS_TRIG01_REG, (1<<OTX_CORE2CORE_GPIONO));
  // here we have registered to interrupts being generated 
  // during going from low to hight
  return 0;
}

int DspCore2CoreInterruptsDisable(void)
{
  return 0;
}

int DspCore2CoreEnable()
{
  g_pBuffersState = OTX_ASP_BUFFERS_STATE_ADDR;
  //CHNG: 2 0 3 1
  g_pBuffersState->nOffsets[2] = 0;
  g_pBuffersState->nOffsets[0] = g_cbBufSize;
  g_pBuffersState->nOffsets[1] = g_cbBufSize * 2;
  g_pBuffersState->nOffsets[3] = g_cbBufSize * 3;
  //CHNG: sz == g_cbBufSize
  g_pBuffersState->coBufPartSize = g_cbBufSize;
  //CHNG: flags == zero-fill
  g_pBuffersState->dwRxTxFlags = 0;

  OsaLinuxRegisterCallback( DspCore2CoreRx, OSA_CALLBACK_DSP, OSA_LINUX_CALLBACK_RX | OSA_LINUX_CALLBACK_NB );
  OsaLinuxRegisterCallback( DspCore2CoreTx, OSA_CALLBACK_DSP, OSA_LINUX_CALLBACK_TX | OSA_LINUX_CALLBACK_NB );

  g_RxDspBufPtr = OTX_EXTEN_RX_BUF_ADDR;
  g_TxDspBufPtr = OTX_EXTEN_TX_BUF_ADDR;
  g_DspBufSize = g_cbBufSize;
  DspCore2CoreInterruptsEnable();

  return 0;
}

int DspCore2CoreDisable()
{
  DspCore2CoreInterruptsDisable();

  OsaLinuxDeleteCallback( DspCore2CoreTx );
  OsaLinuxDeleteCallback( DspCore2CoreRx );

  return 0;
}

// 
int DspCore2CoreInt()
{
  OsaLinuxRunCallBack( OSA_CALLBACK_DSP, 'r', g_RxDspBufPtr, g_DspBufSize );

  return 0;
}

#endif

#endif // DAVINCI
