/* $Workfile: OsaLinuxCallbacks.cpp $ $Revision: 5 $ $Modtime: 10/05/09 9:23a $*/
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 2007, All Rights Reserved

*******************************************************************************/

// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places

// Only include with DaVinci
#ifdef DAVINCI

#include <Sys/Core/OdList.h>
#include "OsaLinuxRtl.h"  // For OSA_DPRINTF
#include "OsaLinuxAsp.h"
#include "OsaLinuxDmp.h"
#include "OsaLinuxCallbacks.h"


/******************************************************************************
******************************************************************************/
class OsaLinuxCallBack
{
public:
  int nMode;
  OsaLinuxCallBackType type;
  OsaLinuxCallBackFunc func;
};

OsaSpinLock *g_OsaLinuxCallBackLock = NULL;

typedef OdList<OsaLinuxCallBack *> CallBackList;
CallBackList *g_pRxCallBack = NULL, *g_pTxCallBack = NULL;

typedef int OsaLinuxRunCallBackMap( OsaLinuxCallBackType type, char method, char* buffer, int size );
OsaLinuxRunCallBackMap OsaLinuxCallBackSimple;
OsaLinuxRunCallBackMap OsaLinuxCallBackMultiple;
OsaLinuxRunCallBackMap *OsaLinuxCallBackHandler;

extern char* g_TxDspBufPtr;
extern char* g_RxDspBufPtr;

CallBackList* RxCallBacks()
{
  if(!g_pRxCallBack) g_pRxCallBack = new CallBackList;
  return g_pRxCallBack;
}


CallBackList* TxCallBacks()
{
  if(!g_pTxCallBack) g_pTxCallBack = new CallBackList;
  return g_pTxCallBack;
}


/******************************************************************************
  OsaLinuxRegisterCallback
******************************************************************************/
int OsaLinuxCallBackTP( OsaLinuxCallBackType type, char on )
{
  static u32 all_types = 0;
  static u32 types_num = 0;
  const u32 f = 1 << (type+1);
  if ((all_types&f)==0 && on) {
    all_types |= f;
    types_num++;
  } else if ((all_types&f)!=0 && !on) {
    all_types &= ~f;
    types_num--;
  } else {
    return 0;
  }
  if( types_num <= 2 ) {
    OsaLinuxCallBackHandler = OsaLinuxCallBackSimple;
  } else {
    OsaLinuxCallBackHandler = OsaLinuxCallBackMultiple;
  }
  return 0;
}

/******************************************************************************
  OsaLinuxRegisterCallback
******************************************************************************/
int OsaLinuxRegisterCallback( OsaLinuxCallBackFunc func, OsaLinuxCallBackType type, int nMode )
{
  OsaSingleLock localLock( g_OsaLinuxCallBackLock );
  OsaLinuxCallBack *st = new OsaLinuxCallBack;
  if(!st) return -EINVAL;

  st->nMode = nMode;
  st->func = func;
  st->type = type;
  CallBackList* list = 0;
  if( nMode & OSA_LINUX_CALLBACK_RX )
    list = RxCallBacks();
  if( nMode & OSA_LINUX_CALLBACK_TX )
    list = TxCallBacks();
  if( list == 0 ){
    delete st;
    return -EINVAL;
  }
  list->PushBack( st );
  OsaLinuxCallBackTP( type, 1 );
  return 0;
}

/******************************************************************************
  OsaLinuxDeleteCallback
******************************************************************************/
int OsaLinuxDeleteCallback( OsaLinuxCallBackFunc func )
{
  OsaSingleLock localLock( g_OsaLinuxCallBackLock );
  for( int i = 0; i < 2; ++i ) {
    CallBackList*& list = i ? g_pRxCallBack : g_pTxCallBack;
    if( list == 0 )
      continue;
    for( CallBackList::Iterator i = list->Begin(); i != list->End(); ++i )
      if( (**i).func == func ) {
        OsaLinuxCallBackTP( (**i).type, 0 );
        delete *i;
        list->Erase( i );
        if( list->Size() == 0 ) {
          delete list;
          list = 0;
        }
        return 0;
      }
  }
  return -EINVAL;
}

/******************************************************************************
  OsaLinuxCallBackSimple
******************************************************************************/
int OsaLinuxCallBackSimple( OsaLinuxCallBackType type, char method, char* buffer, int size )
{
  int should_stop = 1;
  CallBackList* list = ( method == 't' ) ? g_pTxCallBack : g_pRxCallBack;
  if( list == 0 )
    return should_stop;
  for( CallBackList::Iterator i = list->Begin(); i != list->End(); ++i ) {
    if( (**i).nMode & OSA_LINUX_CALLBACK_NB )
      should_stop = 0;
    (**i).func( 0, 0, buffer, size );
  }

  return should_stop;
}

/******************************************************************************
  OsaLinuxCallBackMultiple
******************************************************************************/
int OsaLinuxCallBackMultiple( OsaLinuxCallBackType type, char method, char* buffer, int size )
{
  int should_stop = 1;
  CallBackList* list = ( method == 't' ) ? g_pTxCallBack : g_pRxCallBack;
  if( list == 0 )
    return should_stop;

  if( type != OSA_CALLBACK_DEV ) {
    return 1;
  }
  CallBackList::Iterator first = list->End();
  CallBackList::Iterator second = list->End();
  for( CallBackList::Iterator i = list->Begin(); i != list->End(); ++i ) {
    if( first == list->End() && (**i).type == OSA_CALLBACK_EXTEN )
      first = i;
    else if( second == list->End() && (**i).type == OSA_CALLBACK_DSP )
      second = i;
  }
  if( first == list->End() || second == list->End() ) {
    return 1;
  }
  if( method == 't' ) {
    static char dspbuffer[ 8 * 1024 ];
    memcpy( g_TxDspBufPtr, dspbuffer, size );
    (**second).func( 0, 0, buffer, size );    // DSP
    (**first).func( 0, 0, dspbuffer, size );  // DAHDI
    //(**first).func( 0, 0, g_TxDspBufPtr, size );      // DAHDI
    //(**second).func( 0, 0, buffer, size );            // DSP
  } else if( method == 'r' ) {
    static char dspbuffer[ 8 * 1024 ];
    memcpy( dspbuffer, g_RxDspBufPtr, size );
    (**second).func( 0, 0, buffer, size );    // DSP
    (**first).func( 0, 0, dspbuffer, size );  // DAHDI
  }
  return 0;
}

int OsaLinuxCallBackMultiple_Old( OsaLinuxCallBackType type, char method, char* buffer, int size )
{
  int should_stop = 1;
  CallBackList* list = ( method == 't' ) ? g_pTxCallBack : g_pRxCallBack;
  if( list == 0 )
    return should_stop;

  OsaLinuxCallBackType need_type = -1;
  if( type == OSA_CALLBACK_DEV && method == 'r' ) {
    need_type = OSA_CALLBACK_DSP;
  } else if( type == OSA_CALLBACK_DSP && method == 'r' ) {
    need_type = OSA_CALLBACK_EXTEN;
  } else if( type == OSA_CALLBACK_DEV && method == 't' ) {
    CallBackList::Iterator first = list->End();
    CallBackList::Iterator second = list->End();
    for( CallBackList::Iterator i = list->Begin(); i != list->End(); ++i ) {
      if( (**i).nMode & OSA_LINUX_CALLBACK_NB )
        should_stop = 0;
      if( first == list->End() && (**i).type == OSA_CALLBACK_EXTEN )
        first = i;
      else if( second == list->End() && (**i).type == OSA_CALLBACK_DSP )
        second = i;
    }
    if( first != list->End() && second != list->End() ) {
      (**first).func( 0, 0, g_TxDspBufPtr, size );
/*
      static int ii = 0;
      ++ii;
      if( ii % 1000 == 0 ) {
        printk( "Data before DSP TX\n" );
        for( int i = 0; i < size; ++i ) {
          if( i % 32 == 0 ) 
            printk( "\n" );
          printk( "%02x ", buffer[ i ] );
        }
        printk( "\n\n" );
      }
*/
      (**second).func( 0, 0, buffer, size );
      return 0;
    } else {
      return 1;
    }
  } else {
    return 1;
  }

  if( type == OSA_CALLBACK_DEV ) {
    static int ii = 0;
    ++ii;
    if( ii % 1000 == 0 ) {
/*
      printk( "Data before DSP:\n" );
      for( int i = 0; i < size; ++i ) {
        if( i % 32 == 0 ) 
          printk( "\n" );
        printk( "%02x ", buffer[ i ] );
      }
      printk( "\n\n" );
*/
    }
  }

  if( type == OSA_CALLBACK_DSP ) {
    static int ii = 0;
    ++ii;
    if( ii % 1000 == 0 ) {
/*
      printk( "Data after DSP:\n" );
      for( int i = 0; i < size; ++i ) {
        if( i % 32 == 0 ) 
          printk( "\n" );
        printk( "%02x ", buffer[ i ] );
      }
      printk( "\n\n" );
*/
    }
  }

  for( CallBackList::Iterator i = list->Begin(); i != list->End(); ++i ) {
    if( (**i).nMode & OSA_LINUX_CALLBACK_NB )
      should_stop = 0;
    if( (**i).type == need_type )
      (**i).func( 0, 0, buffer, size );
  }

  return 0;
}

/******************************************************************************
  OsaLinuxRunCallBack
    method should be 't' (for TX) or 'r' (for RX)
    returns 0 if next process should stop
******************************************************************************/
int OsaLinuxRunCallBack( OsaLinuxCallBackType type, char method, char* buffer, int size )
{
  if( g_OsaLinuxCallBackLock && OsaLinuxCallBackHandler ) {
    OsaSingleLock localLock( g_OsaLinuxCallBackLock );
    return OsaLinuxCallBackHandler( type, method, buffer, size );
  }
  return 1;
}

/******************************************************************************
  OsaLinuxCallbackEnable
******************************************************************************/
int OsaLinuxCallbackEnable( OsaLinuxCallBackType type )
{
  if( g_OsaLinuxCallBackLock == 0 ) {
    g_OsaLinuxCallBackLock = new OsaSpinLock;
  }
  OsaLinuxCallBackTP( type, 1 );
  return 0;
}

/******************************************************************************
  OsaLinuxCallbackDisable
******************************************************************************/
int OsaLinuxCallbackDisable( OsaLinuxCallBackType type )
{
  OsaLinuxCallBackTP( type, 0 );
  if( g_OsaLinuxCallBackLock == 0 )
    return 0;
  //delete g_OsaLinuxCallBackLock;
  //g_OsaLinuxCallBackLock = 0;
  return 0;
}

#endif // DAVINCI

