/* $Workfile: OsaLinuxDaVinciI2c.c $   $Revision: 2 $    $Modtime: 1/22/09 1:08p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  Supports communication with I2C devices (specific for Alvis-CSI).

  Device      Address
  ==========  ============
  MSP430      0x23  (0x40)
  EEprom      0x50
  IOExpander  0x20

  @normal Copyright (C) Odin TeleSystems Inc., 2008-2009, All Rights Reserved

*******************************************************************************/


// Only include with DaVinci
#ifdef DAVINCI


#include <linux/config.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/rtc.h>
#include <linux/bcd.h>
#include <asm/rtc.h>
#include <asm/arch/i2c-client.h>


// DBG is not defined for C files only CPP
#if OSA_DBG
#define DBG 1
#endif
#include "OsaLinuxRtl.h"  // For OSA_DPRINTF
#if OSA_DBG
#undef DBG
#endif

#include "OsaLinuxDaVinciI2c.h"

//
// I2C Message Identifiers
//
#define OD_SET_RTC_PARAMS       0
#define OD_GET_RTC_PARAMS       1
#define OD_SEND_IR_DATA         2   
#define OD_GET_LAST_IR_VALUE    3
#define OD_GET_INPUT_STATE      4
#define OD_GET_EVENT            5
#define OD_SET_OUTPUT_STATE     6
#define OD_GET_VERSION          7
#define OD_SLEEP_MODE           8
#define OD_GET_TEMPERATURE      9
#define SET_TEMP_LIMIT         10
#define RESET_MAX_TEMP         11

OsaInt32 OsaLinuxDaVinciRtcGetVersion(OsaUint8 *nRtcFwVersion)
{
	char rtcdata [3];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = OD_GET_VERSION; // Command index
	rc = davinci_i2c_write (2, rtcdata, 0x23);

  if (!rc) {
    udelay (100);
    rc = davinci_i2c_read (3, rtcdata, 0x23);
  	udelay (100);
  }
	
  *nRtcFwVersion = rtcdata[2];

  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcGetTemperature(OsaInt8 *currentDegC, OsaInt8 *limitDegC, OsaInt8 *maxDegC)
{
	char rtcdata [5];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = OD_GET_TEMPERATURE; // Command index
	rc = davinci_i2c_write (2, rtcdata, 0x23);

  if (!rc) {
    udelay (100);
    rc = davinci_i2c_read (5, rtcdata, 0x23);
  	udelay (100);
  }
	
  *currentDegC = (OsaInt8) rtcdata[2];
  *limitDegC   = (OsaInt8) rtcdata[3];
  *maxDegC     = (OsaInt8) rtcdata[4];

  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcResetMaxTemp(void)
{
	char rtcdata [5];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = RESET_MAX_TEMP; // Command index
	rc = davinci_i2c_write (2, rtcdata, 0x23);

	udelay (100);
	
  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcSetTempLimit(OsaInt8 limitDegC)
{
	char rtcdata [5];
  OsaInt32 rc;

	rtcdata [0] = 3; // Message length
	rtcdata [1] = SET_TEMP_LIMIT; // Command index
	rtcdata [2] = limitDegC; // If the temp exceeds this limit we will get interrupts from the RTC
	rc = davinci_i2c_write (3, rtcdata, 0x23);

	udelay (100);
	
  return rc;
}	


OsaInt32 OsaLinuxDaVinciRtcSetOutputState(OsaUint8 p2Val, OsaUint8 p3Val)
{
	char rtcdata [4];
  OsaInt32 rc;

	rtcdata [0] = 4; // Message length
	rtcdata [1] = OD_SET_OUTPUT_STATE; // Command index
	rtcdata [2] = p2Val; // Port2 value
	rtcdata [3] = p3Val; // Port3 value
	rc = davinci_i2c_write (4, rtcdata, 0x23);

	udelay (100);

  return rc;
}	

OsaInt32 OsaLinuxDaVinciRtcGetInputState(OsaUint8 *p2Val, OsaUint8 *p3Val)
{
	char rtcdata [4];
  OsaInt32 rc;

	rtcdata [0] = 2; // Message length
	rtcdata [1] = OD_GET_INPUT_STATE; // Command index
	rc = davinci_i2c_write (2, rtcdata, 0x23);

  if (!rc) {
    udelay (100);
    rc = davinci_i2c_read (4, rtcdata, 0x23);
    udelay (100);
  }

  *p2Val = (OsaUint8) rtcdata[2];
  *p3Val = (OsaUint8) rtcdata[3];

  return rc;
}	

OsaInt32 OsaLinuxDaVinciEepromReadBuf(OsaUint16 nAddress, OsaUint8 *pBuf, OsaInt32 cbBuf)
{
  OsaInt32 rc;
	char data[2];
  
  data[0] = (char) ((nAddress >> 8) & 0xFF);
  data[1] = (char) ((nAddress >> 0) & 0xFF);

  //OSA_DPRINTF(("In OsaLinuxDaVinciEepromReadBuf()\n"));

	rc = davinci_i2c_write (2, data, 0x50);
  if (!rc) {
	  rc = davinci_i2c_read (cbBuf, (char *)pBuf, 0x50);
  }

  return rc;
}

OsaInt32 OsaLinuxDaVinciEepromWritePage(OsaUint16 nAddress, OsaUint8 *pBuf, OsaInt32 cbBuf)
{
  OsaInt32 rc;
	char data[64+2] = { 0x7f, 0, 0xFF}; // 64 for the maximum page size, and 2 for the address
  int  i = 0;

  //OSA_DPRINTF(("In OsaLinuxDaVinciEepromWriteBuf()\n"));

  if (cbBuf > 64) {
    OSA_EPRINTF(("OsaLinuxDaVinciEepromWritePage() Buffer larger than page size. Must be 64 bytes or less.\n"));
    return -1000;
  }

  data[0] = (char) ((nAddress >> 8) & 0xFF);
  data[1] = (char) ((nAddress >> 0) & 0xFF);

  for (i = 0; i < cbBuf; i++) {
    data[i+2] = (char)pBuf[i];
  }

	rc = davinci_i2c_write (cbBuf + 2, data, 0x50);
  udelay (100);
  
  return rc;
}

OsaInt32 OsaLinuxDaVinciI2cRead(OsaInt32 cbBuf, OsaUint8 *pBuf, OsaUint8 nI2cAddress)
{
	return davinci_i2c_read(cbBuf, (char *)pBuf, nI2cAddress);
}

OsaInt32 OsaLinuxDaVinciI2cWrite(OsaInt32 cbBuf, OsaUint8 *pBuf, OsaUint8 nI2cAddress)
{
	return davinci_i2c_write(cbBuf, (char *)pBuf, nI2cAddress);
}


#endif // DAVINCI
