/* $Workfile: OsaLinuxTdmoe.c $   $Revision: 4 $    $Modtime: 2/09/10 11:45a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 2007, All Rights Reserved

*******************************************************************************/

/******************************************************************************
  
  Driver exported entry/exit functions.

  Some of the Linux kernel headers don't like to be compiled with g++ (C++)
  so we need a file that is pure C. This file has two purposes: it implements
  the two Linux module specific entry exit routines, init_module and
  cleanup_module, the file operation callbacks, and also some routines that's
  needed for proc operations.

******************************************************************************/
// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places

// Only include with DaVinci
#ifdef DAVINCI


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/notifier.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>


// DBG is not defined for C files only CPP
#if OSA_DBG
#define DBG 1
#endif
#include "OsaLinuxRtl.h"  // For OSA_DPRINTF
#if OSA_DBG
#undef DBG
#endif

#include "OsaLinuxCallbacks.h"
#include "OsaLinuxTdmoe.h"
#include "OsaLinuxDmp.h"
#include "DspCore2Core.h"


#include <asm/arch/hardware.h>
#include <asm/arch/edma.h>
#include <asm/arch/mcbsp.h>

#define TDMOE_BUFFERS 2
#define ETH_P_TDMOE 0xd00d
#define TDMOE_PACKET_SIZE 1024
#define TDMOE_FRAME_SIZE  128
#define TDMOE_DEFMAC "\xff\xff\xff\xff\xff\xff"

typedef enum {
  OSA_TDMOE_STATE_UNINITIALIZED, 
  OSA_TDMOE_STATE_INITIALIZED, 
  OSA_TDMOE_STATE_OPENED, 
  OSA_TDMOE_STATE_ENABLED, 
} OsaTdmoeStateT;

struct TdmoeHeader {
  unsigned char frames;
  //unsigned short seq;
  unsigned char seq_lo;
  unsigned char seq_hi;
};

static OsaTdmoeStateT g_eDeviceState = OSA_TDMOE_STATE_UNINITIALIZED;
static int g_cbBufSize;
static void *g_pCallbackContext = 0;
static void (*g_pCallbackFuncTxIrq)(unsigned long, void *) = 0;
static void (*g_pCallbackFuncRxIrq)(unsigned long, void *) = 0;
static char* OsaTdmoeRxBuffer[TDMOE_BUFFERS];
static char* OsaTdmoeTxBuffer;
static int OsaTdmoeRxBuf;
static struct net_device *OsaTdmoeDevice;
static char OsaTdmoeAddr[ETH_ALEN];
static char OsaTdmoeRemoteAddr[ETH_ALEN];
static char* OsaTdmoeRxRef;
static int OsaTdmoeRxFilled;
static char OsaTdmoeNextSeqBroken;
static int OsaTdmoeFrames;


// Statistics:
static unsigned int cnt_rx;
static unsigned int cnt_rx_broken;
static unsigned int cnt_rx_unseq;
static unsigned int cnt_rx_unlen;
static unsigned int cnt_rx_overrun;
static unsigned int cnt_tx;


ssize_t OsaLinuxTdmoeWrite(const char *buffer, size_t count);
ssize_t OsaLinuxTdmoeRead(char *buffer, size_t count);


static int ProcessData(void)
{
  int next = 0;
  OsaTdmoeRxRef = OsaTdmoeRxBuffer[OsaTdmoeRxBuf];
  next = OsaLinuxRunCallBack(OSA_CALLBACK_DEV, 'r', OsaTdmoeRxRef,g_cbBufSize);
  if (next!=0) {
    int nStatus = OSA_ASP_DMA_COMPLETE;
    if(OsaTdmoeRxBuf == 0) {
      nStatus |= OSA_ASP_DMA_PING_BUF_READY;
    } else if (OsaTdmoeRxBuf == 1) {
      nStatus |= OSA_ASP_DMA_PONG_BUF_READY;
    }
    g_pCallbackFuncRxIrq(nStatus,g_pCallbackContext);
    g_pCallbackFuncTxIrq(nStatus,g_pCallbackContext);
  } else {
    OsaLinuxRunCallBack(OSA_CALLBACK_DEV, 'r', OsaTdmoeTxBuffer,g_cbBufSize);
    OsaLinuxTdmoeWrite(OsaTdmoeTxBuffer,g_cbBufSize);
  }
}

static int OsaLinuxTdmoeTransmit(const char* msg, int size)
{
  static unsigned short seq = 0;
  struct sk_buff *skb;
  struct net_device *dev;
  unsigned char* addr;
  unsigned char* header;

  dev = OsaTdmoeDevice;
  addr = OsaTdmoeAddr;

  if (!dev) {
    dev = dev_get_by_name("eth0");
    addr = "\0\0\0\0\0\0";
  }

  skb = dev_alloc_skb(size + sizeof(struct TdmoeHeader) + dev->hard_header_len);
  if (!skb) {
    return -ENOMEM;
  }
  skb_reserve(skb, dev->hard_header_len);

  header = skb_put(skb,sizeof(struct TdmoeHeader));
  header[0] = OsaTdmoeFrames;
  header[1] = seq >> 8;
  header[2] = seq & 0xff;
  seq++;

  memcpy(skb_put(skb,size), msg, size);

  skb->protocol = __constant_htons(ETH_P_TDMOE);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,22)
  skb_set_network_header(skb,0);
#else
  skb->nh.raw = skb->data;
#endif
  skb->dev = dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24)
  dev_hard_header(skb, dev, ETH_P_TDMOE, addr, dev->dev_addr, skb->len);
#else
  if (dev->hard_header)
    dev->hard_header(skb, dev, ETH_P_TDMOE, addr, dev->dev_addr, skb->len);
#endif
#if 1
  dev_queue_xmit(skb);
#else
  kfree_skb(skb);
#endif
  cnt_tx++;
  return 0;
}

#define OSA_DPRINTF_ONCE(msg) \
  { static int once = 0; if(once == 0) OSA_DPRINTF(msg) ; once = 1; }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,14)
static int OsaLinuxTdmoeReceive(struct sk_buff *skb, struct net_device *dev, struct packet_type *pt, 
                                struct net_device *orig_dev)
#else
static int OsaLinuxTdmoeReceive(struct sk_buff *skb, struct net_device *dev, struct packet_type *pt)
#endif
{
  unsigned char frames = 0;
  unsigned short seq_num = 0;
  static unsigned short last_seq = 0;
  if (g_eDeviceState != OSA_TDMOE_STATE_ENABLED) {
    OSA_DPRINTF_ONCE(("OsaLinuxTdmoeReceive(): Erorr: packet has come while TDMoE is disabled\n"));
    kfree_skb(skb);
    return 0;
  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
  if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0) {
#else
  if (skb_is_nonlinear(skb) && skb_linearize(skb, GFP_ATOMIC) != 0) {
#endif
    kfree_skb(skb);
    return 0;
  }

  if (memcmp(OsaTdmoeRemoteAddr,TDMOE_DEFMAC,ETH_ALEN)!=0) {
    if (memcmp(OsaTdmoeRemoteAddr,skb->mac.raw + ETH_ALEN,ETH_ALEN) != 0) {
      kfree_skb(skb);
      return 0;
    }
  }

  cnt_rx++;

  if (skb->len >= sizeof(struct TdmoeHeader)) {
    frames = * skb->data;
    seq_num = (skb->data[1]<<8) | (skb->data[2]);
    OsaTdmoeFrames = frames;
  } else {
    cnt_rx_broken++;
    kfree_skb(skb);
    OSA_DPRINTF_ONCE(("OsaLinuxTdmoeReceive(): Error: packet has no header\n"));
    return 0;
  }

  if (seq_num != last_seq + 1 && !OsaTdmoeNextSeqBroken) {
    cnt_rx_unseq++;
    OSA_DPRINTF_ONCE(("OsaLinuxTdmoeReceive(): Warning: seqence is not correct\n"));
  }
  OsaTdmoeNextSeqBroken = 0;
  last_seq = seq_num;

  if ((skb->len - sizeof(struct TdmoeHeader))%TDMOE_FRAME_SIZE != 0) {
    cnt_rx_unlen++;
    kfree_skb(skb);
    OSA_DPRINTF_ONCE(("OsaLinuxTdmoeReceive(): Error: packet size is not correct\n"));
    return 0;
  }

  memcpy(OsaTdmoeRxBuffer[OsaTdmoeRxBuf]+OsaTdmoeRxFilled, skb->data + sizeof(struct TdmoeHeader), skb->len - sizeof(struct TdmoeHeader));
  OsaTdmoeRxFilled += skb->len - sizeof(struct TdmoeHeader);
  if (OsaTdmoeRxFilled == g_cbBufSize) {
    if (OsaTdmoeRxRef) {
      cnt_rx_overrun++;
      OsaTdmoeRxRef = 0;
    }
    ProcessData();
    OsaTdmoeRxFilled = 0;
    OsaTdmoeRxBuf++;
    OsaTdmoeRxBuf %= TDMOE_BUFFERS;
    OsaTdmoeDevice = skb->dev;
    memcpy(OsaTdmoeAddr, skb->mac.raw + ETH_ALEN, sizeof(OsaTdmoeAddr));
  } else if (OsaTdmoeRxFilled > g_cbBufSize) {
    OSA_DPRINTF(("OsaLinuxTdmoeReceive(): Error: Received packet does not fit the buffer\n"));
    OsaTdmoeRxFilled = 0;
    cnt_rx_unlen++;
  }
  kfree_skb(skb);
  return 0;
}


/******************************************************************************
  OsaLinuxTdmoeGetReadPtr
    return 1 on success, Or Error if negative
******************************************************************************/
ssize_t OsaLinuxTdmoeGetReadPtr(char **pBuffer)
{
  *pBuffer = OsaTdmoeRxRef;
  return 0;
}


static struct packet_type OsaLinuxTdmoePacketType = {
	.type = __constant_htons(ETH_P_TDMOE),
	.dev = NULL,
	.func = OsaLinuxTdmoeReceive,
};

/******************************************************************************
  OsaLinuxTdmoeInit
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxTdmoeInit(int cbBufSize, void *pContext, void (*funcServeTxIrq)(unsigned long, void *), void (*funcServeRxIrq)(unsigned long, void *))
{
  int i;

  printk("Hello: OsaLinuxTdmoeInit()\n");
  if (g_eDeviceState != OSA_TDMOE_STATE_UNINITIALIZED) {
    OSA_DPRINTF(("OsaLinuxTdmoeInit(): Erorr: device is already initialized\n"));
    return -EINVAL;
  }

  g_cbBufSize = cbBufSize;
  g_pCallbackContext   = pContext;
  g_pCallbackFuncTxIrq = funcServeTxIrq;
  g_pCallbackFuncRxIrq = funcServeRxIrq;
  OsaTdmoeRxRef = 0;

  for (i = 0; i<TDMOE_BUFFERS; i++) {
    OsaTdmoeRxBuffer[i] = 0;
  }
  OsaTdmoeRxBuf = 0;
  OsaTdmoeNextSeqBroken = 1;

  memset(OsaTdmoeRemoteAddr,'\xff',sizeof(OsaTdmoeRemoteAddr));

  g_eDeviceState = OSA_TDMOE_STATE_INITIALIZED;

  OsaLinuxCallbackEnable(OSA_CALLBACK_DEV);

  return 0;
}


/******************************************************************************
  OsaLinuxTdmoeFreeBuffers
    return void
******************************************************************************/
static void OsaLinuxTdmoeFreeBuffers(void)
{
  int i;

  for (i=0; i<TDMOE_BUFFERS; i++) {
    kfree(OsaTdmoeRxBuffer[i]);
    OsaTdmoeRxBuffer[i] = 0;
  }

  kfree(OsaTdmoeTxBuffer);
  OsaTdmoeTxBuffer = 0;
}


/******************************************************************************
  OsaLinuxTdmoeAllocBuffers
    return 1 on success
******************************************************************************/
int OsaLinuxTdmoeAllocBuffers(int cbBufSize)
{
  int i;

/*
  if (g_eDeviceState != OSA_TDMOE_STATE_OPENED) {
    OSA_EPRINTF(("OsaLinuxTdmoeAllocBuffers(): The buffer size can not be changed in this state.\n"));
    OsaLinuxTdmoeFreeBuffers();
    return -EINVAL;
  }
*/

  OsaLinuxTdmoeFreeBuffers();
  g_cbBufSize = cbBufSize;

  for (i=0; i<TDMOE_BUFFERS; i++) {
    OsaTdmoeRxBuffer[i] = kmalloc(g_cbBufSize, GFP_KERNEL);
    if (OsaTdmoeRxBuffer[i] == 0) {
      return 0;
    }
  }

  OsaTdmoeTxBuffer = kmalloc(g_cbBufSize, GFP_KERNEL);
  if (OsaTdmoeTxBuffer == 0) {
    return 0;
  }

  return 1;
}


/******************************************************************************
  OsaLinuxTdmoeOpen
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxTdmoeOpen(void) 
{
  if (g_eDeviceState != OSA_TDMOE_STATE_INITIALIZED) {
    OSA_DPRINTF(("OsaLinuxTdmoeOpen(): device is not in the initalized state\n"));
    return -EEXIST;
  }

  if(!OsaLinuxTdmoeAllocBuffers(g_cbBufSize)) {
    OSA_DPRINTF(("OsaLinuxTdmoeOpen(): error allocating buffers\n"));
    return -ENOMEM;
  }

  dev_add_pack(&OsaLinuxTdmoePacketType);

  g_eDeviceState = OSA_TDMOE_STATE_OPENED;

  return 0;
}


/******************************************************************************
  OsaLinuxTdmoeClose
    return void
******************************************************************************/
void OsaLinuxTdmoeClose(void)
{
  OsaLinuxCallbackDisable(OSA_CALLBACK_DEV);

  dev_remove_pack(&OsaLinuxTdmoePacketType);

  if (g_eDeviceState == OSA_TDMOE_STATE_ENABLED) {
    OsaLinuxTdmoeDisable(); // Changes the state to OSA_ASP_STATE_OPENED
  }

  if (g_eDeviceState == OSA_TDMOE_STATE_OPENED) {
    OsaLinuxTdmoeFreeBuffers();
    g_eDeviceState = OSA_TDMOE_STATE_INITIALIZED;
  }
}


/******************************************************************************
  OsaLinuxTdmoeEnable
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxTdmoeEnable(int bStartRx, int bStartTx) 
{
  if (g_eDeviceState != OSA_TDMOE_STATE_OPENED){
    OSA_DPRINTF(("OsaLinuxTdmoeEnable(): device is not in its opened state\n"));
    return -EEXIST;
  }

  g_eDeviceState = OSA_TDMOE_STATE_ENABLED; // Everything is running

  return 0; // Success
}


/******************************************************************************
  OsaLinuxTdmoeDisable
    return 0->OK, Or Error
******************************************************************************/
int OsaLinuxTdmoeDisable(void) 
{
  if (g_eDeviceState != OSA_TDMOE_STATE_ENABLED){
    OSA_DPRINTF(("OsaLinuxTdmoeDisable: device is not in its enabled state\n"));
    return -EEXIST;
  }

  g_eDeviceState = OSA_TDMOE_STATE_OPENED; // Back to the previous state

  return 0;

}


/******************************************************************************
  OsaLinuxTdmoeWrite
    return count, Or Error if negative
******************************************************************************/
ssize_t OsaLinuxTdmoeWrite(const char *buffer, size_t count)
{
  int i;

  if (g_eDeviceState != OSA_TDMOE_STATE_ENABLED) {
    return -ENOMEM;
  }
  for (i=count; i>0;) {
    size_t peace = OsaTdmoeFrames * TDMOE_FRAME_SIZE;
    if (peace > i) {
      peace = i;
    }
    OsaLinuxTdmoeTransmit(buffer,peace);
    buffer += peace;
    i -= peace;
  }
  return count; // Success
}


/******************************************************************************
  OsaLinuxTdmoeRead
    return count, Or Error if negative
******************************************************************************/
ssize_t OsaLinuxTdmoeRead(char *buffer, size_t count)
{
  if (g_eDeviceState != OSA_TDMOE_STATE_ENABLED) {
    return -ENOMEM;
  }
  if (OsaTdmoeRxRef) {
    memcpy(buffer, OsaTdmoeRxRef, count);
  } else {
    memset(buffer, 0x00, count);
  }
  return count; // Success
}


/******************************************************************************
  OsaLinuxTdmoeSetMacAddress
    return 0, if success
******************************************************************************/
int OsaLinuxTdmoeSetMacAddress(const unsigned char* addr)
{
  memcpy(OsaTdmoeRemoteAddr,addr,sizeof(OsaTdmoeRemoteAddr));
  return 0;
}


/******************************************************************************
  OsaLinuxTdmoeGetMaccAddress
    return binary remote mac address
******************************************************************************/
extern const unsigned char* OsaLinuxTdmoeGetMacAddress(void)
{
  return OsaTdmoeRemoteAddr;
}


static struct proc_dir_entry* tdmoe_proc;


/******************************************************************************
  OsaLinuxTdmoeCreateProc
    gives the data to the user
******************************************************************************/
static ssize_t tdmoe_proc_read(char *buffer, char **start, off_t offset, int size, int *eof, void *data)
{
  int i = 0;
  i += snprintf(buffer+i, size-i, "rx count  %d\n", cnt_rx);
  i += snprintf(buffer+i, size-i, "rx broken %d\n", cnt_rx_broken);
  i += snprintf(buffer+i, size-i, "rx seq    %d\n", cnt_rx_unseq);
  i += snprintf(buffer+i, size-i, "rx len    %d\n", cnt_rx_unlen);
  i += snprintf(buffer+i, size-i, "tx count  %d\n", cnt_tx);
  return i;
}


/******************************************************************************
  tdmoe_proc_write
    takes data from the user
******************************************************************************/
static ssize_t tdmoe_proc_write(struct file* file, const char* buffer, unsigned long count, void *data)
{
  return count;
}


/******************************************************************************
  OsaLinuxTdmoeCreateProc
    creates proc files inside of given directory
******************************************************************************/
extern int OsaLinuxTdmoeCreateProc(struct proc_dir_entry* proc)
{
  tdmoe_proc = create_proc_entry("tdmoe", 0644, proc);
  if (tdmoe_proc != NULL) {
    tdmoe_proc->owner = THIS_MODULE;
    tdmoe_proc->read_proc = tdmoe_proc_read;
    tdmoe_proc->write_proc = tdmoe_proc_write;
  }
  return 0;
}


/******************************************************************************
  OsaLinuxTdmoeDestroyProc
    destroys proc files inside of given directory
******************************************************************************/
extern int OsaLinuxTdmoeDestroyProc(struct proc_dir_entry* proc)
{
  remove_proc_entry("tdmoe", proc);
  return 0;
}

#endif // DAVINCI
